using sys.io.File;
using sys.io.FileOutput;

class PerfMeter {

    private var iterCount: Int = 0 ;

    /** Timestamp, in secs of the first iteration. **/
    private var firstIterTime: Float = 0.0;

    private var iterTimeStart: Float;

    private var updateInterval: Float ;

    private var prefix: String ;

    private var writeToFile: Bool;

    public function new(prefix: String, updateInterval: Float = 1.0, writeToFile: Bool = true) {
        this.prefix = prefix;
        this.updateInterval = updateInterval;
        this.writeToFile = writeToFile;
    }

    public function iterStart(): Void {
        // Take time of iteration start
        this.iterTimeStart = Sys.time() ;

        if (this.firstIterTime == 0.0) {
            this.firstIterTime = this.iterTimeStart ;
        }

    }

    private var timeAccu: Float = 0.0;
    private var meanAccu: Float = 0.0;
    private var varianceAccu: Float = 0.0;


    private function _accumulate(sample: Float): Void {
        /* Sample R code
        accumulate <- function(table, row, mean_col, var_col, iteration, sample) {
        delta = sample - table[row, mean_col]
        table[row, mean_col] = table[row, mean_col] + delta / iteration
        table[row, var_col] = table[row, var_col] + delta * (sample - table[row, mean_col])
        return(table)
        }
        */

        this.timeAccu += sample ;
        var delta:Float = sample - this.meanAccu ;
        this.meanAccu = this.meanAccu + delta / this.iterCount ;
        this.varianceAccu = this.varianceAccu + delta * (sample - this.meanAccu) ;
    }

    public function iterEnd(): Void {
        var now:Float = Sys.time() ;
        // Accumulates time / mean / var
        var iterDuration: Float = now - this.iterTimeStart ;

        // Increase iteration counter
        this.iterCount +=1 ;

        // accumulation.matrix <- accumulate(table=accumulation.matrix, row=segment, mean_col = "mle.rel", var_col = "mle.rel.var", iteration = simulationStep, sample = mle.rel)
        this._accumulate(iterDuration);

        var elapsed: Float = now - this.firstIterTime ;
        if(elapsed > this.updateInterval) {

            // accumulation.matrix[,"mle.rel.var"] <- accumulation.matrix[,"mle.rel.var"] / (numberOfSimulations - 1)
            this.varianceAccu = this.varianceAccu / (this.iterCount-1) ;

            var SD: Float = Math.sqrt(this.varianceAccu) ;

            var mean:Float = this.timeAccu / this.iterCount ;
            var summary: String = "PerfMeter " + this.prefix + " Report: Iterations=" + this.iterCount + "\tTotal time=" + this.timeAccu + "\tMean (gross)=" + mean + "\tMean (accu)=" + this.meanAccu + "\tVar=" + this.varianceAccu + "\tSD=" + SD ;
            trace(summary) ;

            if(this.writeToFile) {
                var filename: String = "PerfMeter-"+this.prefix+"-"+now+".txt" ;
                // File.saveContent(filename, summary) ;
                var fout:FileOutput = File.write(filename, false);
                fout.writeString(summary);
                fout.flush();
                fout.close();
            }

            this.firstIterTime = 0 ;
            this.iterCount = 0 ;
            this.timeAccu = 0 ;
            this.meanAccu = 0 ;
            this.varianceAccu = 0 ;

        }

    }

    public static function main() {
        var simulationTime: Float = 30.0;

        var pm: PerfMeter = new PerfMeter("LocalTest", 5) ;

        var startTime: Float = Sys.time() ;

        while(true) {
            pm.iterStart();

            // Sys.sleep(0.1) ;
            Sys.sleep(0.1 + Math.random() * 0.8) ;

            var after: Float = Sys.time() ;
            pm.iterEnd() ;

            if( Sys.time() - startTime > simulationTime) {
                break;
            }
        }
    }

}