# import libraries
import httplib2  # for http requests
import re  # for regular expressions
import os

import bge  # for blender game engine
import time  # for time operations

import aud  # for playing audio

#
# Whether to use the hand-written or the Haxe-transpiled version
from haxe_lib import MaryTTSBlendSequencer
#from marytts_manual import MaryTTSBlendSequencer


# URL composition instructions at: http://mary.dfki.de:59125/documentation.html

# Select either your local installation or the remote dfki server
MARY_TTS_ADDRESS = "decad.sb.dfki.de:59125"
# MARY_TTS_ADDRESS = "mary.dfki.de:59125"
#MARY_TTS_ADDRESS = "localhost:59125"

# Select the voice (it has to be installed in the server)

MaryTTSVoice_DFKI_PRUDENCE = "dfki-prudence"
MaryTTSVoice_DFKI_PRUDENCE_HSMM = "dfki-prudence-hsmm"
MaryTTSVoice_DFKI_POPPY = "dfki-poppy"
MaryTTSVoice_DFKI_POPPY_HSMM = "dfki-poppy-hsmm"
MaryTTSVoice_CMU_SLT = "cmu-slt"
MaryTTSVoice_CMU_SLT_HSMM = "cmu-slt-hsmm"

# This is the best voice:
MARY_TTS_VOICE = MaryTTSVoice_DFKI_PRUDENCE
# For the localhost use this:
#MARY_TTS_VOICE = MaryTTSVoice_CMU_SLT_HSMM


# Compose final URL
MARY_TTS_HTTP_ADDRESS = "http://"+MARY_TTS_ADDRESS
MARY_TTS_AUDIO_PARAMETER = "&OUTPUT_TYPE=AUDIO&AUDIO=WAVE_FILE&LOCALE=en_US"  # "&INPUT_TYPE=TEXT&OUTPUT_TYPE=AUDIO&AUDIO=WAVE_FILE&LOCALE=en_US&VOICE="+MARY_TTS_VOICE
MARY_TTS_REALISED_DURATION_PARAMETER = "&OUTPUT_TYPE=REALISED_DURATIONS&LOCALE=en_US"  # "&INPUT_TYPE=TEXT&OUTPUT_TYPE=REALISED_DURATIONS&LOCALE=en_US&VOICE="+MARY_TTS_VOICE
MARY_TTS_VOICE_PARAMETER = "&VOICE=" + MARY_TTS_VOICE

#
# INITIALIZE
#
co = bge.logic.getCurrentController()
obj = co.owner
scene = obj.scene
obj = scene.objects["Anna"]  # Take the reference to the armature



#
#  GLOBAL VARIABELS
scene_dir = bge.logic.expandPath('//')
# print("CWD: '{}'. Scene directory: '{}'. ".format(os.getcwd(), scene_dir))
sequencer = MaryTTSBlendSequencer()  # The parser and sequencer
viseme_weights = [0.0] * MaryTTSBlendSequencer.get_viseme_count()  # The always up-to-date vector of weights
audio_device = aud.device()  # The default audio device to playback the sound
last_audio_handle = None  # set to the handle of the last played sample


last_speech_start = 0.0
SPEEC_INTERVAL = 5.0


# This function updates the mouth of the character
def update():
    global viseme_weights
    global last_speech_start

    #
    # Update the vector of weights
    now = time.time()  # current time
    sequencer.update(now, viseme_weights=viseme_weights)
    # print (viseme_weights)

    #
    # Update the position of the driving bones
    for viseme, weight in zip(MaryTTSBlendSequencer.VISEMES, viseme_weights):
        ch = obj.channels['shapekey-'+viseme]
        #ch.location = 0, weight, 0
        sw = - 2.0 * (weight * weight * weight) + 3.0 * (weight * weight)
        ch.location = 0, sw, 0

    # Update the armature to move the bones (and consequently the shapekeys)
    obj.update()

    #
    # Repeat a sentence
    if now - last_speech_start > SPEEC_INTERVAL:
        #ttsResponse("The quick brown fox jumps over the lazy dog.")
        speak_direct()
        last_speech_start = now


# The response from MaryTTS server
def ttsResponse(tts):

    tts = re.sub('\ ', '+', tts)

    #
    # Http request and response for audio file
    print("Getting audio...")
    audio = httplib2.Http()
    request_url = MARY_TTS_HTTP_ADDRESS + "/process?INPUT_TEXT=" + tts + "&INPUT_TYPE=TEXT" + MARY_TTS_VOICE_PARAMETER;
    print("REQUEST URL: "+request_url)

    audio_resp, audio_content = audio.request(request_url+MARY_TTS_AUDIO_PARAMETER, "GET")
    print("--> {} {}".format(audio_resp.status, audio_resp.reason))

    # Write audio file
    with open(os.path.join(scene_dir, 'tts.wav'), 'wb') as f:
        f.write(audio_content)

    #
    # Http request and response for Realised durations
    print("Getting realized durations...")
    duration = httplib2.Http()
    duration_resp, duration_content = duration.request(request_url+MARY_TTS_REALISED_DURATION_PARAMETER, "GET")
    print("--> {} {}".format(duration_resp.status, duration_resp.reason))

    # Write realized duration to text file
    with open(os.path.join(scene_dir, 'realized_duration.txt'), 'wb') as f:
        f.write(duration_content)


def any_just_activated(controller: bge.types.SCA_IController) -> bool:
    """Returns True if any of the sensors attached to this controller has been just activated.
     Useful to filter only events for key pressed and ignore key released.

    :param controller:
    :return: True if any of the sensors has been just activated.
    """

    for s in controller.sensors:
        if s.status == bge.logic.KX_SENSOR_JUST_ACTIVATED:
            return True

    return False


def speak_direct():
    # Given phonemes
    f = open(os.path.join(scene_dir, 'realized_duration.txt'), 'r')

    durations_str = ""
    # storing phonemes and durations from marytts file to arrays
    for line in f:
        durations_str += line

    sequencer.parse_realized_durations(realized_durations=durations_str)
    sequencer.reset_timers()

    # load sound file
    factory = aud.Factory(os.path.join(scene_dir, 'tts.wav'))

    # play the audio, this return a handle to control play/pause
    last_audio_handle = audio_device.play(factory)


def speak():
    global last_audio_handle

    if any_just_activated(bge.logic.getCurrentController()):
        speak_direct()


#
# STOP speaking!
def stop():
    if any_just_activated(bge.logic.getCurrentController()):
        sequencer.stop_sequencer()
        if last_audio_handle is not None:
            last_audio_handle.stop()


#
# PRESET SENTENCES
def scentence1():
    if any_just_activated(bge.logic.getCurrentController()):
        ttsResponse('If everything goes right both individuals shake hands and go back to the world with a smile.')


def scentence2():
    if any_just_activated(bge.logic.getCurrentController()):
        ttsResponse('The quick brown fox jumps over the lazy dog.')


def scentence3():
    if any_just_activated(bge.logic.getCurrentController()):
        ttsResponse('Dont be so gloomy after all its not that awful.')


def scentence4():
    if any_just_activated(bge.logic.getCurrentController()):
        ttsResponse('I am gonna make him an offer he cant refuse.')


def scentence5():
    if any_just_activated(bge.logic.getCurrentController()):
        ttsResponse('They may take our lives but not our freedom.')


def sentence6():
    if any_just_activated(bge.logic.getCurrentController()):
        ttsResponse('A, b, c, d, e, f, g, h, i, j, k, l, m, n, o, p, q, r, s, t, u, v, w, x, y, z.')
        # ttsResponse('A, e, i, o, u.')


def sentence7():
    if any_just_activated(bge.logic.getCurrentController()):
        ttsResponse('0, 1, 2, 3, 4, 5, 6, 7, 8, 9.')
